/*
    Copyright 2004,2005,2006,2007 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <ctype.h>

#ifdef WIN32
    #include <winsock.h>
    #include <malloc.h>
    #include "winerr.h"

    #define close       closesocket
    #define sleep       Sleep
    #define ASYNC       int async = 1; \
                        if(ioctlsocket(sd, FIONBIO, (void *)&async) < 0) std_err();
#else
    #include <unistd.h>
    #include <sys/socket.h>
    #include <sys/types.h>
    #include <arpa/inet.h>
    #include <netinet/in.h>
    #include <netdb.h>
    #include <fcntl.h>

    #define stristr     strcasestr
    #define strnicmp    strncasecmp
    #define ASYNC       int async; \
                        if((async = fcntl (sd, F_GETFL)) < 0 || \
                          fcntl(sd, F_SETFL, async | O_NONBLOCK) < 0) std_err();
#endif



#define VER         "0.2"
#define BUFFSZ      4096
#define TIMEOUT     20



uint8_t *jmpspc(uint8_t *p);
int conn_timeout(int sock);
uint8_t *showhttp80(uint16_t port, uint8_t *port80);
int check(uint8_t *host, uint16_t port, uint8_t *user, uint8_t *pass, uint8_t *uri);
uint8_t *mydown_uri2hex(uint8_t *uri);
void mydown_get_host(uint8_t *url, uint8_t **hostx, uint16_t *portx, uint8_t **urix, uint8_t **userx, uint8_t **passx, int verbose);
uint8_t *mydown_base64_encode(uint8_t *data, int *size);
int timeout(int sock);
uint32_t resolv(char *host);
void std_err(void);



int main(int argc, char *argv[]) {
    FILE        *fd,
                *fdo;
    int         html    = 0,
                len,
                code;
    uint16_t    port;
    uint8_t     *buff,
                *log,
                *listfile,
                *outfile,
                *user,
                *pass,
                *host,
                *url,
                *uri,
                *p;

#ifdef WIN32
    WSADATA    wsadata;
    WSAStartup(MAKEWORD(1,0), &wsadata);
#endif

    setbuf(stdout, NULL);

    fputs("\n"
        "Web passwords and links checker "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stdout);

    if(argc < 3) {
        printf("\n"
            "Usage: %s <list_file.txt> <output_file.TXT/HTML>\n"
            "\n"
            "Note that this tool supports only the classical HTTP server Authorization\n"
            "method, no cookies, POST or others\n"
            "the input file must be in clean text, if you have a html do a copy&paste of\n"
            "the visualized text in a text file\n"
            "if the output file already exists the data will be appended\n"
            "\n", argv[0]);
        exit(1);
    }

    listfile = argv[1];
    outfile  = argv[2];

    printf("- open input list file \"%s\"\n", listfile);
    fd = fopen(listfile, "rb");
    if(!fd) std_err();

    printf("- open output file \"%s\"\n", outfile);
    fdo = fopen(outfile, "ab");
    if(!fdo) std_err();

    if((p = strrchr(outfile, '.')) && !strnicmp(p, ".htm", 4)) {
        printf("- html output activated\n");
        html = 1;
        fprintf(fdo, "<html>\n");
    }

    buff = malloc(BUFFSZ + 1);
    log  = malloc(BUFFSZ + 1);
    if(!buff || !log) std_err();

    printf("- timeout for connections and data receiving is set to %d seconds\n", TIMEOUT);
    printf("- start check:\n\n");

    while(fgets(buff, BUFFSZ, fd)) {
        for(p = buff; *p && ((*p == ' ') || (*p == '\t')); p++);    // skip front spaces
        url = p;
        for(p = url; *p && (*p != '\n') && (*p != '\r'); p++);      // delimit
        *p = 0;
        for(p--; (p >= url) && ((*p == ' ') || (*p == '\t')); p--); // skip end spaces
        *(p + 1) = 0;

        printf("\nTEXT     %s\n", url);

        mydown_get_host(url, &host, &port, &uri, &user, &pass, 1);

        if(!user) {
            user = jmpspc(uri);

            p = strchr(user, ':');
            if(p) {
                *p = 0;
                pass = p + 1;
            } else {
                pass = jmpspc(user);
                jmpspc(pass);           // avoids spaces after the password
            }

            fprintf(stderr,
                "  user   %s\n"
                "  pass   %s\n",
                user,
                pass);

        } else if(!pass) {
            pass = "";
        }

        code = check(host, port, user, pass, uri);

        if((code < 200) || (code > 206)) {
            printf("\n         WRONG HTTP CODE %d\n", code);
        } else {
            printf("\n         VALID!!!\n");

            len = 0;
            if(user && pass) {
                len = snprintf(
                    log,
                    BUFFSZ,
                    "%s:%s@",
                    user, pass);

                if((len < 0) || (len > BUFFSZ)) len = BUFFSZ;
            }

            snprintf(
                log + len,
                BUFFSZ - len,
                "%s%s/%s",
                host, showhttp80(port, NULL), uri);

            if(html) {
                fprintf(fdo,
                    "<a href=\"http://%s\">http://%s</a><br>\n",
                        log,
                        log);
            } else {
                fprintf(fdo,
                    "http://%s\n",
                    log);
            }

            fflush(fdo);
        }

        free(uri);
    }

    if(html) fprintf(fdo, "</html>\n");

    fclose(fd);
    fclose(fdo);
    free(buff);
    free(log);
    printf("\n- finished\n");
    return(0);
}



uint8_t *jmpspc(uint8_t *p) {
    while(*p &&  (*p != ' ') && (*p != '\t'))  p++;
    *p++ = 0;
    while(*p && ((*p == ' ') || (*p == '\t'))) p++;
    return(p);
}



int conn_timeout(int sock) {
    struct  timeval tout;
    fd_set  r,
            w,
            e;
    int     err;

    tout.tv_sec  = TIMEOUT;
    tout.tv_usec = 0;
    FD_ZERO(&r);
    FD_ZERO(&w);
    FD_ZERO(&e);
    FD_SET(sock, &r);
    FD_SET(sock, &w);
    FD_SET(sock, &e);
    err = select(sock + 1, &r, &w, &e, &tout);
    if(err < 0) std_err();
    if((err != 1) || (FD_ISSET(sock, &e))) return(-1);
    return(0);
}



uint8_t *showhttp80(uint16_t port, uint8_t *port80) {
    static uint8_t  mini[7];

    if(!port80) port80 = mini;
    *port80 = 0;
    if(port != 80) sprintf(port80, ":%hu", port);
    return(port80);
}



int check(uint8_t *host, uint16_t port, uint8_t *user, uint8_t *pass, uint8_t *uri) {
    struct  sockaddr_in peer;
    int         sd          = 0,
                len,
                t,
                b64len,
                code        = -1;
    uint8_t     port80[7],
                *query      = NULL,
                *buff       = NULL,
                *b64        = NULL,
                *userpass   = NULL,
                *p;

    showhttp80(port, port80);

    if(user && pass) {
        printf("CHECK    http://%s:%s@%s%s/%s\n",
            user, pass, host, port80, uri);
    } else {
        printf("CHECK    http://%s%s/%s\n",
            host, port80, uri);
    }

    printf("         resolv, ");
    peer.sin_addr.s_addr = resolv(host);
    if(peer.sin_addr.s_addr == INADDR_NONE) goto quit;
    peer.sin_port        = htons(port);
    peer.sin_family      = AF_INET;

    sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if(!sd) std_err();

    printf("connect, ");
    ASYNC;
    connect(sd, (struct sockaddr *)&peer, sizeof(peer));
    if(conn_timeout(sd) < 0) {
        printf("timeout or refused");
        goto quit;
    }

    if(user && pass) {
        userpass = malloc(strlen(user) + 1 + strlen(pass) + 1);
        if(!userpass) goto quit;
        b64len = sprintf(userpass, "%s:%s", user, pass);
        b64 = mydown_base64_encode(userpass, &b64len);
    }

    len =
        200                                       + // my format strings
        strlen(uri)                               +
        strlen(host)                              +
        5                                         +
        (b64             ? strlen(b64)       : 0) +
        1;

    query = malloc(len);
    if(!query) std_err();

    len = sprintf(
        query,
        "GET /%s HTTP/1.1\r\n"
        "Host: %s%s\r\n"
        "Connection: close\r\n",
        uri,
        host, port80);

#define HTTP_APPEND  len += sprintf(query + len,

    if(b64) {
        HTTP_APPEND "Authorization: Basic %s\r\n", b64);
    }
    HTTP_APPEND "\r\n");

#undef HTTP_APPEND  

    printf("send request, ");
    if(send(sd, query, len, 0) < 0) goto quit;

    buff = malloc(BUFFSZ + 1);
    if(!buff) std_err();

    printf("receive data, ");
    for(len = 0; len < BUFFSZ; len += t) {
        if(timeout(sd) < 0) {
            printf("timeout (%d sec), ", TIMEOUT);
            break;
        }
        t = recv(sd, buff + len, BUFFSZ - len, 0);
        if(t <= 0) break;
        fputc('.', stdout);
    }
    buff[len] = 0;

    if(len < 15)  goto quit;    // HTTP/1.1 200
    p = strchr(buff, ' ');
    if(!p) goto quit;

    code = atoi(p + 1);
    printf(" code:   %d", code);

quit:
    if(sd)          close(sd);
    if(b64)         free(b64);
    if(buff)        free(buff);
    if(query)       free(query);
    if(userpass)    free(userpass);
    return(code);
}



uint8_t *mydown_uri2hex(uint8_t *uri) {
    static const uint8_t hex[16] = "0123456789abcdef";
    uint8_t     *ret,
                *p,
                c;

    ret = malloc((strlen(uri) * 3) + 1);
    if(!ret) return(NULL);

    for(p = ret; *uri; uri++) {
        c = *uri;
        if(isprint(c)) {
            *p++ = c;
        } else {
            *p++ = '%';
            *p++ = hex[c >> 4];
            *p++ = hex[c & 15];
        }
    }
    *p = 0;

    return(ret);
}



void mydown_get_host(uint8_t *url, uint8_t **hostx, uint16_t *portx, uint8_t **urix, uint8_t **userx, uint8_t **passx, int verbose) {
    uint16_t    port  = 80;
    uint8_t     *host = NULL,
                *uri  = NULL,
                *user = NULL,
                *pass = NULL,
                *p;

    host = url;

    p = strstr(host, "://");    // handle http://
    if(!p) p = strstr(host, ":\\\\");
    if(p) {
        for(p += 3; *p; p++) {  // in case of http:////
            if((*p != '/') && (*p != '\\')) break;
        }
        host = p;
    }

    for(p = host; *p; p++) {    // search the uri
        if((*p == '/') || (*p == '\\')) {
            uri = p;
            break;
        }
    }
    if(uri) {
        *uri++ = 0;
        uri = mydown_uri2hex(uri);
    }
    if(!uri) uri = strdup("");  // in case mydown_uri2hex fails

    p = strchr(host, '@');
    if(p) {
        *p = 0;

        user = host;

        pass = strchr(host, ':');
        if(pass) {
            *pass++ = 0;
        } else {
            pass = "";
        }

        host = p + 1;
    }

    p = strchr(host, ':');
    if(p) {
        *p = 0;
        port = atoi(p + 1);
    }

    if(verbose >= 0) fprintf(stderr, "  %s\n", url);
    if(verbose > 0) fprintf(stderr,
        "  host   %s : %hu\n"
        "  uri    %s\n",
        host, port,
        uri);

    if(user) {
        if(verbose > 0) fprintf(stderr,
            "  user   %s\n"
            "  pass   %s\n",
            user,
            pass);
    }

    *hostx = host;
    *portx = port;
    *urix  = uri;
    *userx = user;
    *passx = pass;
}



uint8_t *mydown_base64_encode(uint8_t *data, int *size) {
    int         len;
    uint8_t     *buff,
                *p,
                a,
                b,
                c;
    static const uint8_t base[64] = {
        'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P',
        'Q','R','S','T','U','V','W','X','Y','Z','a','b','c','d','e','f',
        'g','h','i','j','k','l','m','n','o','p','q','r','s','t','u','v',
        'w','x','y','z','0','1','2','3','4','5','6','7','8','9','+','/'
    };

    if(!size || (*size < 0)) {      // use size -1 for auto text size!
        len = strlen(data);
    } else {
        len = *size;
    }
    buff = malloc(((len / 3) << 2) + 6);
    if(!buff) return(NULL);

    p = buff;
    do {
        a     = data[0];
        b     = data[1];
        c     = data[2];
        *p++  = base[(a >> 2) & 63];
        *p++  = base[(((a &  3) << 4) | ((b >> 4) & 15)) & 63];
        *p++  = base[(((b & 15) << 2) | ((c >> 6) &  3)) & 63];
        *p++  = base[c & 63];
        data += 3;
        len  -= 3;
    } while(len > 0);
    *p = 0;

    for(; len < 0; len++) *(p + len) = '=';

    if(size) *size = p - buff;
    return(buff);
}



int timeout(int sock) {
    struct  timeval tout;
    fd_set  fd_read;
    int     err;

    tout.tv_sec = TIMEOUT;
    tout.tv_usec = 0;
    FD_ZERO(&fd_read);
    FD_SET(sock, &fd_read);
    err = select(sock + 1, &fd_read, NULL, NULL, &tout);
    if(err < 0) std_err();
    if(!err) return(-1);
    return(0);
}



uint32_t resolv(char *host) {
    struct      hostent *hp;
    uint32_t    host_ip;

    host_ip = inet_addr(host);
    if(host_ip == INADDR_NONE) {
        hp = gethostbyname(host);
        if(!hp) {
            printf("\nError: Unable to resolve hostname (%s)\n", host);
            return(INADDR_NONE);
        } else host_ip = *(uint32_t *)(hp->h_addr);
    }

    return(host_ip);
}



#ifndef WIN32
    void std_err(void) {
        perror("\nError");
        exit(1);
    }
#endif


